"""Contains geometry functions for translating between representations and coordinate systems
"""

import json
import logging
from typing import Optional
import osgeo
import yaml
from osgeo import ogr
from osgeo import osr


def calculate_centroid_from_wkt(wkt: str) -> tuple:
    """Given WKT, return lat/lon of centroid.
    Arguments:
        wkt: Well Known Text (WKT) format of geometry
    returns:
        Tuple of (lat, lon) representing centroid of the geometry
    """

    # Convert to geometry instance
    loc_geom = ogr.CreateGeometryFromWkt(wkt)
    return loc_geom.Centroid().GetX(), loc_geom.Centroid().GetY()


def calculate_overlap_percent(check_bounds: str, bounding_box: str) -> float:
    """Calculates and returns the percentage overlap between the two boundaries.
       The calculation determines the overlap shape between the two parameters and
       then calculates the percentage by dividing the overlap area by the bounding
       box area, and returns that value.
    Arguments:
        check_bounds: GeoJSON of boundary to check
        bounding_box: GeoJSON of boundary to check against
    Return:
        The calculated overlap percent (0.0 - 1.0) or 0.0 if there is no overlap.
        If an exception is detected, a warning message is logged and 0.0 is returned.
    """
    try:
        check_poly = ogr.CreateGeometryFromJson(str(check_bounds))
        bbox_poly = ogr.CreateGeometryFromJson(str(bounding_box))

        if check_poly and bbox_poly:
            intersection = bbox_poly.Intersection(check_poly)
            if intersection:
                return intersection.Area() / check_poly.Area()
    except Exception as ex:
        logging.warning("Exception caught while calculating shape overlap: %s", str(ex))

    return 0.0


def convert_geometry(geometry: ogr.Geometry, new_spatialreference: osr.SpatialReference) -> ogr.Geometry:
    """Converts the geometry to the new spatial reference if possible
    Arguments:
        geometry - The geometry to transform
        new_spatialreference - The spatial reference to change to
    Returns:
        The transformed geometry or the original geometry. If either the
        new Spatial Reference parameter is None, or the geometry doesn't
        have a spatial reference, then the original geometry is returned.
    """
    if not new_spatialreference or not geometry:
        return geometry

    return_geometry = geometry
    try:
        geom_sr = geometry.GetSpatialReference()
        if int(osgeo.__version__[0]) >= 3:
            # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
            # pylint: disable=no-member
            geom_sr.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
        if geom_sr and not new_spatialreference.IsSame(geom_sr):
            transform = osr.CreateCoordinateTransformation(geom_sr, new_spatialreference)
            new_geom = geometry.Clone()
            if new_geom:
                new_geom.Transform(transform)
                return_geometry = new_geom
    except Exception as ex:
        logging.warning("Exception caught while transforming geometries: %s", str(ex))
        logging.warning("    Returning original geometry")

    return return_geometry


def geojson_to_tuples(bounding_box: str) -> tuple:
    """Returns the bounds of the shape
    Arguments:
        bounding_box: the GeoJSON of the geometry
    Return:
        A tuple containing the bounds in (min Y, max Y, min X, max X) order
    """
    yaml_geom = yaml.safe_load(bounding_box)
    current_geom = ogr.CreateGeometryFromJson(json.dumps(yaml_geom))
    return geometry_to_tuples(current_geom)


def geometry_to_tuples(geom: ogr.Geometry) -> tuple:
    """Returns the bounds of the shape
    Arguments:
        geom: the geometry to return the bounds of
    Return:
        A tuple containing the bounds in (min Y, max Y, min X, max X) order
    """
    current_env = geom.GetEnvelope()

    return current_env[2], current_env[3], current_env[0], current_env[1]


def geometry_to_geojson(geom: ogr.Geometry, alt_coord_type: str = None, alt_coord_code: str = None) -> str:
    """Converts a geometry to geojson.
    Args:
        geom: The geometry to convert to JSON
        alt_coord_type: the alternate geographic coordinate system type if geometry doesn't have one defined
        alt_coord_code: the alternate geographic coordinate system associated with the type
    Returns:
        The geojson string for the geometry
    Note:
        If the geometry doesn't have a spatial reference associated with it, both the default
        coordinate system type and code must be specified for a coordinate system to be assigned to
        the returning JSON. The original geometry is left unaltered.
    """
    ref_sys = geom.GetSpatialReference()
    geom_json = json.loads(geom.ExportToJson())
    if not ref_sys:
        if alt_coord_type and alt_coord_code:
            # No coordinate system, use what was passed in
            geom_json['crs'] = {'type': str(alt_coord_type), 'properties': {'code': str(alt_coord_code)}}
    else:
        # Use the existing coordinate system to inform the GeoJSON
        geom_json['crs'] = {
            'type': ref_sys.GetAttrValue("AUTHORITY", 0),
            'properties': {
                'code': ref_sys.GetAttrValue("AUTHORITY", 1)
            }
        }
    return json.dumps(geom_json)


def polygon_from_ring(ring: ogr.Geometry, epsg: int = None) -> Optional[ogr.Geometry]:
    """Creates a polygon from the linear ring geometry passed in
    Arguments:
        ring: the linear ring to create the polygon with
        epsg: the EPSG code to assign to the polygon
    Return:
        The created polygon, or None if an EPSG code is specified and can't be loaded
    Exceptions:
        Raises a RuntimeError if a SRID is specified but can't be loaded
    """
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)

    ref_sys = osr.SpatialReference()
    if epsg is not None:
        if ref_sys.ImportFromEPSG(int(epsg)) == ogr.OGRERR_NONE:
            poly.AssignSpatialReference(ref_sys)
        else:
            return None
    return poly



"""CheckMD Class"""

from typing import List, NamedTuple, Optional, TextIO


# pylint: disable=invalid-name
class CheckMD(NamedTuple):
    """This is the CheckMD class based off of NamedTuple which can be used in
    order to store data passed into argparse."""
    timestamp: str
    season: str
    experiment: str
    working_folder: str
    list_files: List[TextIO]
    container_name: Optional[str] = None
    target_container_name: Optional[str] = None
    trigger_name: Optional[str] = None
    context_md: Optional[str] = None

    def get_list_files(self):
        """Returns list_files"""
        return self.list_files



"""Functions for handling LAS,LAZ files
"""

import json
import os
from typing import Optional
import logging
import subprocess
from osgeo import ogr

from agpypeline import geometries


def clip_las(las_path: str, clip_tuple: tuple, out_path: str) -> None:
    """Clip LAS file to polygon.
    Arguments:
      las_path: path to point cloud file
      clip_tuple: tuple containing (minX, maxX, minY, maxY) of clip bounds
      out_path: output file to write
    Notes:
        The clip_tuple is assumed to be in the correct coordinate system for the point cloud file
    """
    bounds_str = "([%s, %s], [%s, %s])" % (clip_tuple[0], clip_tuple[1], clip_tuple[2], clip_tuple[3])

    pdal_dtm = out_path.replace(".las", "_dtm.json")
    with open(pdal_dtm, 'w', encoding='utf-8') as dtm:
        dtm_data = """{
            "pipeline": [
                "%s",
                {
                    "type": "filters.crop",
                    "bounds": "%s"
                },
                {
                    "type": "writers.las",
                    "filename": "%s"
                }
            ]
        }""" % (las_path, bounds_str, out_path)
        logging.debug("Writing dtm file contents: %s", str(dtm_data))
        dtm.write(dtm_data)

    cmd = 'pdal pipeline "%s"' % pdal_dtm
    logging.debug("Running pipeline command: %s", cmd)
    subprocess.call([cmd], shell=True)
    os.remove(pdal_dtm)


def get_las_epsg(file_path: str, json_result: dict = None) -> Optional[str]:
    """Returns the found EPSG code from the LAS header
    Arguments:
        file_path: the path to the file from which to extract the epsg
        json_result: the json from calling pdal info on a .las file
    Return:
        Returns the SRID as a string if found, None is returned otherwise
    """
    epsg = None
    try:
        if json_result is not None:
            keys = list(json_result['stats']['bbox'].keys())
        else:
            stats = subprocess.run(['pdal', 'info', file_path], stderr=subprocess.PIPE,
                                   stdout=subprocess.PIPE, check=True)
            stats_json = json.loads(stats.stdout.decode())
            keys = list(stats_json['stats']['bbox'].keys())
        for key in keys:
            if 'EPSG' in key:
                epsg = key.split(':')[1]
    except Exception as ex:
        logging.debug("Unable to find EPSG in LAS file header")
        logging.debug("    exception caught: %s", str(ex))

    return epsg


def get_las_extents(file_path: str, default_epsg: int = None) -> Optional[str]:
    """Calculate the extent of the given las file and return as GeoJSON.
    Arguments:
        file_path: path to the file from which to load the bounds
        default_epsg: the default EPSG to assume if a file has a boundary but not a coordinate system
    Return:
        Returns the JSON representing the image boundary, or None if the
        bounds could not be loaded
    Notes:
        If a file doesn't have a coordinate system and a default epsg is specified, the
        return JSON will use the default_epsg.
        If a file doesn't have a coordinate system and there isn't a default epsg specified, the boundary
        of the image is not returned (None) and a warning is logged.
    """
    # Get the bounds and the EPSG code
    stats = subprocess.run(['pdal', 'info', file_path], stderr=subprocess.PIPE,
                           stdout=subprocess.PIPE, check=True)
    json_result = json.loads(stats.stdout.decode())
    bounds = json_result['stats']['bbox']['native']['bbox']
    epsg = get_las_epsg(file_path, json_result)
    if epsg is None:
        if default_epsg is not None:
            epsg = default_epsg
        else:
            logging.warning("Unable to find EPSG and not default is specified for file '%s'", file_path)
            return None

    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(bounds['minx'], bounds['maxy'])  # Upper left
    ring.AddPoint(bounds['maxx'], bounds['maxy'])  # Upper right
    ring.AddPoint(bounds['maxx'], bounds['miny'])  # lower right
    ring.AddPoint(bounds['minx'], bounds['miny'])  # lower left
    ring.AddPoint(bounds['minx'], bounds['maxy'])  # Closing the polygon

    poly = geometries.polygon_from_ring(ring, int(epsg))
    if poly:
        return geometries.geometry_to_geojson(poly)

    logging.error('Failed to create bounding polygon with EPSG %s from las file "%s"', str(epsg), file_path)
    return None



#!/usr/bin/env python3
"""
Test algorithm class
"""
from agpypeline.algorithm import Algorithm
from agpypeline import entrypoint

from agpypeline.checkmd import CheckMD
from agpypeline.configuration import Configuration
from agpypeline.environment import Environment


class BasicAlgorithm(Algorithm):
    """Test Algorithm For Integration Tests"""

    # pylint: disable=no-self-use
    def perform_process(self, environment: Environment, check_md: CheckMD, transformer_md: dict,
                        full_md: list) -> dict:
        """Perform the processing of data
        Arguments:
            environment: instance of Environment class
            check_md: metadata for this Transformer execution run
            transformer_md: transformer specific information from previous runs
            full_md: the list of loaded metadata
         """
        return 0


if __name__ == "__main__":
    CONFIGURATION = Configuration()
    entrypoint.entrypoint(CONFIGURATION, BasicAlgorithm())



"""Template class for a Transformer Algorithm
"""
from agpypeline.environment import Environment, CheckMD


class Algorithm:
    """Class for containing a Transformer Algorithm
    """

    def __init__(self):
        """Initializes a class instance
        """

    def perform_process(self, environment: Environment, check_md: CheckMD, transformer_md: dict,
                        full_md: list) -> dict:
        """Perform the processing of data
        Arguments:
            environment: instance of Environment class
            check_md: metadata for this Transformer execution run
            transformer_md: transformer specific information from previous runs
            full_md: the list of loaded metadata
        """
        # pylint: disable=unused-argument,no-self-use
        raise RuntimeError("The Algorithm class method perform_process() must be overridden by a derived class")



#!/usr/bin/env python3
"""Suite of Integration tests
"""
import os
import random
import string

from subprocess import getstatusoutput

SOURCE_FILE = 'integration_tests/basic_algorithm.py'
SOURCE_PATH = os.path.abspath(os.path.join('', SOURCE_FILE))

TESTING_JSON_FILE_PATH = os.path.realpath('./integration_tests/integration_tests_data')
META = os.path.abspath(os.path.join(TESTING_JSON_FILE_PATH, 'meta.yaml'))
INPUT1 = os.path.abspath(os.path.join(TESTING_JSON_FILE_PATH, 'rgb_17_7_W.tif'))


def random_string():
    """generate a random string"""
    k = random.randint(5, 10)
    return ''.join(random.choices(string.ascii_letters + string.digits, k=k))


def test_no_metadata_or_files():
    """Test the output when no metadata or file_list is included on the command line"""
    ret_val, out = getstatusoutput(f'{SOURCE_PATH}')
    assert int(ret_val) == 0 and int(out) == 0


def test_metadata_no_files():
    """Test the output when no file_list is included on the command line"""
    ret_val, out = getstatusoutput(f'{SOURCE_PATH} --metadata {META}')
    assert int(ret_val) == 0 and int(out) == 0


def test_metadata_and_files():
    """Tests the output when valid arguments are entered on the command line"""
    out_dir = random_string()
    os.makedirs(out_dir)
    # This ought not be necessary as the program *should*
    # create it; for now, we'll create the output dir.

    ret_val, out = getstatusoutput(f'{SOURCE_PATH} {INPUT1} --working_space {out_dir} --metadata {META}')
    assert int(ret_val) == 0 and int(out) == 0
